import {ProductSystemGenerator} from "./ProductSystemGenerator";
import {ProductGeneratorResult} from "./ProductGeneratorResult";
import {TemplateElement} from "../../templates/TemplateElement";
import {ProductTemplateSystem} from "../../templates/productGamesystem/ProductTemplateSystem";
import {ProductState} from "../states/ProductState";
import {Transition} from "../transitions/Transition";
import {ProductTransition} from "../transitions/ProductTransition";
import {State} from "../states/State";
import {SimpleTemplateTransition} from "../../templates/simpleGamesystem/SimpleTemplateTransition";
import {state, transition} from "@angular/animations";
import {ScriptAccountCondition} from "../conditions/ScriptAccountCondition";
import {SimpleTemplateState} from "../../templates/simpleGamesystem/SimpleTemplateState";
import {Character} from "../../characters/Character";
import {CharacterRelation} from "../../characters/CharacterRelation";
import {ProductGenerationData} from "./ProductGenerationData";

export class TemplateProductSystemGenerator extends ProductSystemGenerator {

  templateElement: TemplateElement


  constructor(productGamesystem: ProductTemplateSystem, templateElement: TemplateElement) {
    super(productGamesystem);
    this.templateElement = templateElement;
  }

  protected assignGeneratedStatesAndTransitions(generationResult: ProductGeneratorResult) {
    const productTemplateSystem = this.productGamesystem as ProductTemplateSystem
    productTemplateSystem.transitionMap.set(this.templateElement, generationResult.transitions)
    productTemplateSystem.stateMap.set(this.templateElement, generationResult.states)

    console.log("Symmetric", (this.productGamesystem as ProductTemplateSystem).symmetric)
    console.log("Template: ", generationResult.transitions.length)
  }

  protected generateFromBinaryChildsystems(leftSystemData: ProductGenerationData, rightSystemData: ProductGenerationData): ProductGeneratorResult {
    const generatedProductStates: ProductState[] = []
    const generatedProductTransitions: ProductTransition[] = []

    leftSystemData.states.forEach(leftState => {
      rightSystemData.states.forEach(rightState => {
        for(let i=0; i<leftState.outgoingTransitions.length; i++) {
          for(let j=0; j<rightState.outgoingTransitions.length; j++) {
            const startingState = this.generateBinaryProductState(leftState, rightState, generatedProductStates);

            if(startingState != undefined) {

              if(!(this.productGamesystem as ProductTemplateSystem).symmetric) {
                const endingState_right = this.generateBinaryProductState(leftState, rightState.outgoingTransitions[j].endingState, generatedProductStates);
                if(endingState_right != undefined) {
                  this.generateBinaryProductTransition(startingState, endingState_right, rightState.outgoingTransitions[j], generatedProductTransitions, false);
                }

                const endingState_left = this.generateBinaryProductState(leftState.outgoingTransitions[i].endingState, rightState, generatedProductStates);
                if(endingState_left != undefined) {
                  this.generateBinaryProductTransition(startingState, endingState_left, leftState.outgoingTransitions[i], generatedProductTransitions, true)
                }
              }

              const endingState_left_right = this.generateBinaryProductState(leftState.outgoingTransitions[i].endingState, rightState.outgoingTransitions[j].endingState, generatedProductStates);
              if(endingState_left_right != undefined) {
                if((this.productGamesystem as ProductTemplateSystem).symmetric) {
                  if(leftState.equals(rightState) && leftState.outgoingTransitions[i].endingState.equals(rightState.outgoingTransitions[j].endingState)) {
                    this.generateBinaryProductTransitionMulti(startingState, endingState_left_right, leftState.outgoingTransitions[i], rightState.outgoingTransitions[j], generatedProductTransitions);
                  } else {
                    console.log(startingState , endingState_left_right)
                  }
                } else {
                  this.generateBinaryProductTransitionMulti(startingState, endingState_left_right, leftState.outgoingTransitions[i], rightState.outgoingTransitions[j], generatedProductTransitions);
                }

              }
            }
          }

          if(rightState.outgoingTransitions.length == 0 && !(this.productGamesystem as ProductTemplateSystem).symmetric) {
            const startingState = this.generateBinaryProductState(leftState, rightState, generatedProductStates);
            const endingState = this.generateBinaryProductState(leftState.outgoingTransitions[i].endingState, rightState, generatedProductStates);

            if(startingState != undefined && endingState != undefined) {
              this.generateBinaryProductTransition(startingState, endingState, leftState.outgoingTransitions[i], generatedProductTransitions, true)

            }
          }
        }

        if(leftState.outgoingTransitions.length == 0 && !(this.productGamesystem as ProductTemplateSystem).symmetric) {
          for(let j=0; j<rightState.outgoingTransitions.length; j++) {
            const startingState = this.generateBinaryProductState(leftState, rightState, generatedProductStates);
            const endingState = this.generateBinaryProductState(leftState, rightState.outgoingTransitions[j].endingState, generatedProductStates);

            if(startingState != undefined && endingState != undefined) {
              this.generateBinaryProductTransition(startingState, endingState, rightState.outgoingTransitions[j], generatedProductTransitions, false);

            }

          }
        }
      })
    })

    return new ProductGeneratorResult(generatedProductStates, generatedProductTransitions);
  }

  protected getTransitionConditions(transition: Transition<any>, leftSystem: boolean) {
    const templateElement = this.determineTemplateElement(leftSystem)!;
    if(transition instanceof SimpleTemplateTransition && transition.conditionMap.has(templateElement)) {
      return transition.conditionMap.get(templateElement)!
    }
    return transition.scriptAccountConditions;
  }

  protected getTransitionActions(transition: Transition<any>, leftSystem: boolean) {
    const templateElement = this.determineTemplateElement(leftSystem)!;

    if(transition instanceof SimpleTemplateTransition && transition.actionMap.has(templateElement)) {
      return transition.actionMap.get(templateElement)!
    } else {
      return transition.scriptAccountActions;
    }
  }


  protected getStateConditions(state: State<any>, leftSystem: boolean): ScriptAccountCondition[] {
    const templateElement = this.determineTemplateElement(leftSystem)!
    if(state instanceof  SimpleTemplateState && state.conditionMap.has(templateElement)) {
      return state.conditionMap.get(templateElement)!
    } else {
      return state.conditions
    }
  }

  private determineTemplateElement(leftSystem: boolean) {
    if(this.templateElement instanceof Character) {
      return this.templateElement;
    } else if(this.templateElement instanceof CharacterRelation) {
      return leftSystem ? this.templateElement.firstCharacter : this.templateElement.secondCharacter;
    }
  }
}
